import warnings as wn


class _CallAndHandle:
    def __init__(self, handler, func):
        self._handler = handler
        self._func = func

    def __call__(self, *args, **kwargs):
        try:
            return self._func(*args, **kwargs)
        except Exception as e:
            return self._handler(e, self._func, *args, **kwargs)


def handled_by(handler):
    def decorator(func):
        return _CallAndHandle(handler, func)

    return decorator


class FunctionError:
    def __init__(self, e, func, *args, **kwargs):
        self._e = e
        self._func = func
        self._args = args
        self._kwargs = kwargs

    def __call__(self):
        return self._func(*self._args, **self._kwargs)

    def __str__(self):
        string = ""
        string += f"Exception:\n{self._e}\n"
        string += f"Function:\n{self._func}\n"
        string += f"Arguments:\n{self._args}\n"
        string += f"Key-Word Arguments:\n{self._kwargs}\n"
        return string

    def __repr__(self):
        return str(self)

    def __bool__(self):
        return False

    def __eq__(self, other):
        result = type(self._e) == type(other._e)
        result &= self._e.args == other._e.args
        result &= self._func == other._func
        result &= self._args == other._args
        result &= self._kwargs == other._kwargs
        return result


def print_warning_return_function_error(e, func, *args, **kwargs):
    wn.warn(
        f"Warning: the function {func.__name__} raised a {type(e)} exception when given arguments {args} and kwargs {kwargs}"
    )
    return FunctionError(e, func, *args, **kwargs)



from collections.abc import MutableMapping
import itertools as it
import operator as op

import numpy as np
import xarray as xr

from .defs import UFUNC_TO_OP
from .utils import (
    validate_ntable_init,
    str_ntable,
    xarray_coords_to_dict,
    ttype_to_attrs,
    call,
    setitem,
    validate_engine,
    validate_ttype,
    concatenate_ntables,
    full,
    NULL,
    NTableStopIteration,
    get_method_and_call,
    full_like,
)
from .tabularization import tabularize
from .engines import StandardEngine, ProcessEngine, ThreadEngine
from .handling import (
    handled_by,
    FunctionError,
    print_warning_return_function_error,
)
from .structure import NTableStructure
from .filtering import NTableFilter, contains, matches
from .alchemy import NTableAlchemy, NTableMapAlchemy


class NTableMap(MutableMapping):
    """
    A dictionary-like interface for working with a NTable object along a
    specific dimension.

    Parameters
    ----------
    dim : str
        The dimension to create a ntable map for.

    Raises
    ------
    ValueError
        Raised if the given dimension does not exist.
    """

    def __init__(self, ntable, dim):
        self._ntable = ntable
        self._dim = dim

    @property
    def ntable(self):
        return self._ntable

    @property
    def dim(self):
        return self._dim

    @property
    def alchemy(self):
        return NTableMapAlchemy(self)

    def __dir__(self):
        result = list(self.__dict__)
        result.extend(list(self.keys()))
        return result

    # def _ipython_key_completions_(self, incomplete_key):
    #     if not isinstance(incomplete_key, str):
    #         return []
    #     coords_dict = xarray_coords_to_dict(self._ntable.struct.coords)
    #     dim_coords = coords_dict[self._dim]
    #     result = []
    #     for coord in dim_coords:
    #         if isinstance(coord, str):
    #             if coord.startswith(incomplete_key):
    #                 result.append(coord)
    #
    #     return [coord for coord in dim_coords if coord.startswith(incomplete_key)]

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(f"{attr} is not an attribute of NTable")

    def __str__(self):
        return str(dict(**self))

    def __repr__(self):
        return str(self)

    def __getitem__(self, key):
        return self._ntable.struct.loc[{self._dim: key}]

    def __setitem__(self, key, value):
        if isinstance(key, list):
            if isinstance(value, NTable):
                if len(key) != len(value.ntable_map(self._dim)):
                    raise ValueError(
                        f"value.{self._dim} must have same length as key"
                    )
                for k, v in zip(key, value.ntable_map(self.dim).values()):
                    self[k] = v
            else:
                for k in key:
                    self[k] = value
        else:
            try:
                self._ntable.struct.loc[{self._dim: key}] = value
            except KeyError:
                sample_ntable = next(value for value in self.values())
                sample_coords = xarray_coords_to_dict(
                    sample_ntable.struct.coords
                )
                sample_coords[self._dim][0] = key
                extension = full(
                    NULL(), sample_coords, self._ntable.struct.dims
                )
                intermediate = concatenate_ntables(
                    (self._ntable, extension), dim=self.dim
                )
                intermediate.ntable_map(self.dim)[key] = value
                self._ntable._reflist = intermediate.reflist
                self._ntable._refmap = intermediate.refmap

    def __delitem__(self, key):
        raise NotImplementedError

    def __iter__(self):
        coords_dict = xarray_coords_to_dict(self._ntable.struct.coords)
        return iter(coords_dict[self._dim])

    def __len__(self):
        coords_dict = xarray_coords_to_dict(self._ntable.struct.coords)
        return len(coords_dict[self._dim])

    def contains(self, string):
        return self._ntable.filter[{self._dim: contains(string)}]

    def matches(self, pattern):
        return self._ntable.filter[{self._dim: matches(pattern)}]

    def relabel(self, **kwargs):
        return self._ntable.struct.relabel(**{self._dim: kwargs})

class TabularizedMethod:
    """
    Represents a method whose call should be tabularized.
    """

    def __init__(self, ntbl, method_name):
        self._ntbl = ntbl
        self._method_name = method_name

    def __call__(self,*args, **kwargs):
        handled_ = handled_by(FunctionError)(get_method_and_call)
        return tabularize(engine=self._ntbl.engine)(handled_)(self._ntbl, self._method_name, *args, **kwargs)

    def __str__(self):
        method_name_ntbl = full_like(self._method_name, self._ntbl, lite=True)
        return str(method_name_ntbl)

    def __repr__(self):
        return str(self)


class TabularizedAttributes:
    """
    Used to interface with the attributes of the ELEMENTS of a NTable object
    as opposed to the attributes of the NTable object itself.
    """

    def __init__(self, ntable):
        super().__setattr__("_ntable", ntable)

    def __getattr__(self, attr):
        handled_ = handled_by(FunctionError)(getattr)
        return tabularize(engine=self._ntable.engine)(handled_)(
            self._ntable, attr
        )

    def __setattr__(self, attr, value):
        handled_ = handled_by(FunctionError)(setattr)
        tabularize(engine=self._ntable.engine)(handled_)(
            self._ntable, attr, value
        )


def _next(iterator):
    # this function exists to raise a different type of exception
    # than StopIteration so that the tabularization proces doesn't
    # continue after StopIteration is raised (since it wont be
    # handled until the end of the end of the tabularized call) and
    # issues arise when trying to complete the tabularized call.
    # Throwing an error here prevents the process from finishing.
    try:
        next_item = next(iterator)
        return next_item
    except StopIteration:
        raise NTableStopIteration


class _NTable_Iterator:
    def __init__(self, ntbl):
        self._ntbl = ntbl

    def __next__(self):
        try:
            if isinstance(self._ntbl.engine, ProcessEngine):
                # if the engine is a ProcessEngine, then the StopIteration will
                # be masked by the processes and cannot stop the iteration. As
                # such, the tabularized next will be done with threads.
                engine = ThreadEngine(self._ntbl.engine.processes)
            else:
                engine = self._ntbl.engine
            return tabularize(engine=engine)(_next)(self._ntbl)
        except NTableStopIteration:
            raise StopIteration


class NTable(np.lib.mixins.NDArrayOperatorsMixin):
    """
    N-dimensional, tabular representation of a set of heterogeneous
    python data

    Parameters
    ----------
    reflist : list
        List containing the data to represent.
    refmap : DataArray
        A dataarray whose elements correspond to indexes in the reflist.
        Describes the layout of the data.
    engine : callable, optional
        A map-like callable. Must take in a function as the first argument
        and iterables whose elements will be passed into the function.
        If None, a serial-mapping engine will be used.
    ttype : set, optional
        A set of types. These types define what the NTable object MIGHT contain.
        This set is used to determine if a failed getattr operation should be
        tabularized or not. If None, a set of common types is used.
    validate : bool, optional
        Whether or not validate the NTable constructor inputs. Users making NTable
        objects manually should always set this to True. The default is True.

    Methods
    -------


    """

    def __init__(
        self, reflist, refmap, engine=None, ttype=None, validate=True
    ):

        if engine is None:
            engine = StandardEngine()
        orig_ttype = ttype
        if ttype is None:
            ttype = set()
        if validate:
            validate_ntable_init(reflist, refmap, engine, ttype)
        if orig_ttype is None:
            # if ttype was originally None, needed to define it as
            # something that will pass the validation step. Once
            # passed, we can assign it types based on the contents
            # of the intended NTable
            for item in (reflist[i] for i in refmap.values.flat):
                ttype.add(type(item))
        self._reflist = reflist
        self._refmap = refmap
        self._engine = engine
        self._ttype = ttype

    @property
    def reflist(self):
        """The list of data that the NTable object represents"""
        return self._reflist

    @property
    def refmap(self):
        """The dataarray that describes the layout of the data"""
        return self._refmap

    @property
    def struct(self):
        """
        An object that allows the user to work with the structure of a
        NTable object

        """
        return NTableStructure(self)

    @property
    def tattr(self):
        """
        An object where getting and setting attributes on it results in the
        tabularization of the get or set call

        """
        return TabularizedAttributes(self)

    @property
    def engine(self):
        """The engine that gets used for processing"""
        return self._engine

    @engine.setter
    def engine(self, e):
        validate_engine(e)
        self._engine = e

    @property
    def ttype(self):
        """The data types the NTable object MIGHT contain"""
        return self._ttype

    @ttype.setter
    def ttype(self, t):
        validate_ttype(t)
        self._ttype = t

    @property
    def filter(self):
        """An object used for filtering the NTable object."""
        return NTableFilter(self)

    @property
    def alchemy(self):
        """An object used for special types of lookups/operations."""
        return NTableAlchemy(self)

    def __dir__(self):
        result = list(self.__dict__)
        result.extend(self.struct.dims)
        result.extend(ttype_to_attrs(self.ttype))
        return result

    def __getattr__(self, attr):
        try:
            return self.ntable_map(attr)
        except ValueError:
            attr_dict = ttype_to_attrs(self._ttype)
            if attr in attr_dict:
                if callable(attr_dict[attr]):
                    # If the attribute is callable, save on overhead
                    # by returning a TabularizedMethod object instead
                    # of calling tabularized getattr since
                    # TabularizedMethod will reduce on tabularization
                    # overhead.
                    return TabularizedMethod(self, attr)
                return getattr(self.tattr, attr)
            raise AttributeError(f"{attr} is not an attribute of NTable")

    def __str__(self):
        ttype_strings = sorted([type_.__name__ for type_ in self._ttype])
        ttype_str = "|".join(ttype_strings)
        return f"{str_ntable(self)}\n{self.struct.coords}\nEngine:\n{self._engine}\nTtype:\n{ttype_str}"

    def __repr__(self):
        return str(self)

    def __iter__(self):
        ntable_of_iterators = tabularize(engine=self._engine)(iter)(self)
        return _NTable_Iterator(ntable_of_iterators)

    def __getitem__(self, index):
        handled_ = handled_by(FunctionError)(op.getitem)
        return tabularize(engine=self._engine)(handled_)(self, index)

    def __setitem__(self, index, value):
        handled_ = handled_by(print_warning_return_function_error)(setitem)
        result = tabularize(engine=self._engine)(handled_)(self, index, value)

    def __call__(self, *args, **kwargs):
        handled_ = handled_by(FunctionError)(call)
        return tabularize(engine=self._engine)(handled_)(self, *args, **kwargs)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        ufunc = UFUNC_TO_OP.get(ufunc, ufunc)
        handled_ = handled_by(FunctionError)(ufunc)
        return tabularize(engine=self._engine)(handled_)(*inputs, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        handled_ = handled_by(FunctionError)(func)
        return tabularize(engine=self._engine)(handled_)(*args, **kwargs)

    def item(self):
        return self.struct.item()

    def ntable_map(self, dim):
        """
        Returns a ntable map object for the given dimension

        Parameters
        ----------
        dim : str
            The dimension to create a ntable map for.

        Raises
        ------
        ValueError
            Raised if the given dimension does not exist.

        Returns
        -------
        NTableMap
            An object that allows the user the treat a given dimension object
            as a dictionary-like object.

        """
        if dim not in self.struct.dims:
            raise ValueError(f"{dim} dimension does not exist")
        return NTableMap(self, dim)

    def to_dictionary(self, into=None, enforce_nested_typing=True):
        """
        Convert a NTable object into a dictionary

        Parameters
        ----------
        into : MutableMapping, optional
            A dictionary-like object to store elements into. If None,
            a new dictionary is created and returned.

        enforce_nested_typing : bool
            Flag indicating whether or not nested elements of into
            should be made the same type as into. Default is True.

        Returns
        -------
        into : MutableMapping
            The resulting dictionary-like object.

        """
        if into is None:
            into = {}
        dim = self.struct.dims[0]
        for k, v in self.ntable_map(dim).items():
            if isinstance(v, NTable):
                if v.struct.ndim > 0:
                    if enforce_nested_typing:
                        into[k] = type(into)()
                    else:
                        into[k] = {}
                    v.to_dictionary(into[k])
                else:
                    into[k] = v.struct.item()
            else:
                into[k] = v

        return into

    def to_pandas(self, dtype="object"):
        """
        Convert a NTable object into a Pandas object
        (Series or DataFrame depending on the shape of the NTable object.)

        Parameters
        ----------
        dtype : str, np.dtype, optional
            Data type that the resulting Pandas object should be.
            The default is "object".

        Raises
        ------
        ValueError
            Raised when called on a NTable object with more than 2 dimensions.

        Returns
        -------
        Series or DataFrame
            The resulting Pandas object.

        """
        if self.struct.ndim < 3 and self.struct.ndim > 0:
            return self.to_data_array(dtype=dtype).to_pandas()
        else:
            raise ValueError(
                f"Unable to convert {self.struct.ndim} dimensional NTable to pandas object"
            )

    def to_data_array(self, dtype="object"):
        """
        Convert a NTable object into a DataArray

        Parameters
        ----------
        dtype : str, np.dtype, optional
            Data type that the resulting DataArray should be.
            The default is "object".

        Returns
        -------
        DataArray
            The resulting DataArray.

        """
        array = np.empty(self.refmap.size, dtype="object")
        array[:] = list(self.struct.flat)
        array = array.reshape(self.refmap.shape)
        return xr.DataArray(
            array, self.refmap.coords, self.refmap.dims
        ).astype(dtype)



import inspect
from collections.abc import MutableMapping

import h5py
import numpy as np
import xarray as xr

from .defs import PRINTABLE_TYPES
from .engines import Engine


def _dummy_func(x, y):
    return x + y


def validate_engine(engine):
    if isinstance(engine, Engine):
        return
    else:
        raise ValueError("The input must follow the same interface as map")


def validate_ttype(ttype):
    for t in ttype:
        if not isinstance(t, type):
            raise ValueError("Every value in ttype must be of type type")


def validate_ntable_init(reflist, refmap, engine, ttype):
    if not isinstance(refmap, xr.DataArray):
        raise TypeError("refmap must be an xarray DataArray")
    if len(list(refmap.coords.keys())) == 0:
        raise ValueError("Data map must have non empty coordinates")
    if not isinstance(reflist, list):
        raise TypeError("reflist must be a list")

    for item in refmap.data.flat:
        try:
            reflist[item]
        except IndexError:
            raise ValueError(
                "every element of refmap must be a valid index of reflist"
            )

    # if len(np.unique(refmap)) < refmap.values.size:
    #     raise ValueError("Every element of refmap must be unique")

    validate_engine(engine)
    validate_ttype(ttype)


def ttype_to_attrs(ttype):
    attrs = {}
    for type_ in ttype:
        attrs.update({item[0]:item[1] for item in inspect.getmembers(type_)})
    return attrs


def xarray_coords_to_dict(coords, dims=None):
    coords_dict = {}
    if dims is None:
        keys = coords.keys()
    else:
        keys = dims
    for k in keys:
        try:
            coords_dict[k] = list(coords[k].values)
        except TypeError:
            coords_dict[k] = [coords[k].values.item()]
    return coords_dict


def coords_dims_to_shape(coords, dims):
    if isinstance(coords, xr.DataArray):
        coords = xarray_coords_to_dict(coords)
    return tuple(len(coords[d]) for d in dims)


def basic_refmap(coords, dims):
    shape = coords_dims_to_shape(coords, dims)
    size = np.prod(shape)
    refmap = xr.DataArray(
        np.arange(size, dtype="int").reshape(shape), coords, dims
    )
    return refmap


def full(value, coords, dims, engine=None, ttype=None):
    from .ntable import NTable

    shape = coords_dims_to_shape(coords, dims)
    size = int(np.prod(shape))
    dlist = [value] * size
    dmap = xr.DataArray(
        np.arange(size, dtype="int").reshape(shape), coords, dims
    )
    # TODO: Turn off validation once this has been tested thoroughly
    return NTable(dlist, dmap, engine, ttype)

def full_like(value, ntbl, engine=None, ttype=None, lite=False):
    if lite:
        return full_lite(value, ntbl.struct.coords, ntbl.struct.dims, engine=engine, ttype=ttype)
    return full(value, ntbl.struct.coords, ntbl.struct.dims, engine=engine, ttype=ttype)

def full_lite(value, coords, dims, engine=None, ttype=None):
    from .ntable import NTable

    shape = coords_dims_to_shape(coords, dims)
    size = int(np.prod(shape))
    dlist = [value]
    dmap = xr.DataArray(np.zeros(shape, dtype="int"), coords, dims)
    # TODO: Turn off validation once this has been tested thoroughly
    return NTable(dlist, dmap, engine, ttype)


def str_ntable_element(val):
    try:
        return val.__ntable_element__str__()
    except AttributeError:
        if isinstance(val, np.ndarray):
            return f"ndarray,{val.shape},{val.dtype}"
        elif isinstance(val, xr.DataArray):
            return f"data array,{val.shape},{val.dtype}"
        elif type(val) in PRINTABLE_TYPES:
            return str(val)
        elif isinstance(val, str):
            if len(val) < 15:
                return f'"{val}"'
            else:
                return f"{val[:5]}...{val[-5:]}"
        elif isinstance(val, (list, tuple)):
            if len(val) < 5:
                return f"{val}"
            else:
                return f"{val[:2]}...{val[-2:]}"
        else:
            return type(val).__name__


def str_ntable(ntbl):
    from .tabularization import tabularize

    max_rows = 100 + 25
    max_cols = 20 + 5
    max_other = 10 + 3

    dims = ntbl.struct.dims

    reduce_index = {}

    try:
        if len(ntbl.struct.coords[dims[0]]) > max_rows:
            reduce_index[dims[0]] = np.r_[0:50, -50:0]
    except IndexError:
        pass

    try:
        if len(ntbl.struct.coords[dims[1]]) > max_cols:
            reduce_index[dims[0]] = np.r_[0:10, -10:0]
    except IndexError:
        pass

    print_ellipsises = {}

    for dim in dims[2:]:
        if len(ntbl.struct.coords[dim]) > max_other:
            reduce_index[dim] = np.r_[0:5, -5:0]
            print_ellipsises[dim] = True
        else:
            print_ellipsises[dim] = False

    ntbl = ntbl.struct[reduce_index]

    ntbl = tabularize()(str_ntable_element)(ntbl)

    if ntbl.struct.ndim > 2:
        string = ""
        dim = ntbl.struct.dims[-1]
        ntbl_map = ntbl.ntable_map(dim)
        for i, (k, v) in enumerate(ntbl_map.items()):
            sep = "#" * 79
            string += f"{k}:\n\n{str_ntable(v)}\n\n{sep}\n\n"
            if (i == 4) and (print_ellipsises[dim]):
                string += f". . .\n\n{sep}\n\n"

    elif ntbl.struct.ndim == 0:
        string = str(ntbl.struct.item())
    else:
        string = str(ntbl.to_pandas())
    return string


def any_ntables(iterable):
    from .ntable import NTable

    return any(isinstance(item, NTable) for item in iterable)


def call_args_kwargs(func, *args):
    return func(*args[0], **args[1])


def call(func, *args, **kwargs):
    return func(*args, **kwargs)

def get_method_and_call(object, methodname, *args, **kwargs):
    method = getattr(object, methodname)
    return method(*args, **kwargs)

def setitem(obj, index, value):
    obj[index] = value


class _Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(
                *args, **kwargs
            )
        return cls._instances[cls]


# class SUPREMUM(metaclass=_Singleton):
#     def __gt__(self, other):
#         return True
#
#     def __ge__(self, other):
#         return True
#
#     def __lt__(self, other):
#         return False
#
#     def __le__(self, other):
#         return False

# class INFIMUM(metaclass=_Singleton):
#     def __gt__(self, other):
#         return False
#
#     def __ge__(self, other):
#         return False
#
#     def __lt__(self, other):
#         return True
#
#     def __le__(self, other):
#         return True


class NULL(np.lib.mixins.NDArrayOperatorsMixin, metaclass=_Singleton):
    def __str__(self):
        return "NULL"

    def __repr__(self):
        return "NULL"

    def __ntable_element__str__(self):
        return "NULL"

    def __getitem__(self, index):
        return NULL()

    def __call__(self, *args, **kwargs):
        return NULL()

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        return NULL()

    # def __array_function__(self, func, types, args, kwargs):
    #     return NULL()


def handle_improper_broadcast(bdata, bmap):
    bdata.append(NULL())
    new_bmap_array = np.nan_to_num(bmap, nan=len(bdata) - 1).astype("int")
    new_bmap = xr.DataArray(new_bmap_array, bmap.coords, bmap.dims)
    return bdata, new_bmap


def concatenate_ntables(objs, dim, coords=None):
    from .ntable import NTable

    dmaps = tuple(obj.refmap for obj in objs)
    dlists = tuple(obj.reflist for obj in objs)
    true_sizes = tuple(len(dlist) for dlist in dlists)

    new_engine = objs[0].engine

    new_ttype = set()
    for obj in objs:
        new_ttype |= obj.ttype

    new_dmaps = tuple(
        dmap + sum(true_sizes[0:i]) for i, dmap in enumerate(dmaps)
    )

    # TODO: handle when dim is new and assign new coordinates
    new_dmap = xr.concat(new_dmaps, dim)
    dim_is_new = True
    for new_dmap_ in new_dmaps:
        if dim in new_dmap_.dims:
            dim_is_new = False
            break

    if dim_is_new:
        # if dim is new, then dim was added along axis=0 per xarray.concate
        # for NTables we want the behavior so that it is a new dimension
        # created at the end so we need to rearange...
        new_dmap = new_dmap.transpose(
            *(new_dmap.dims[1:] + (new_dmap.dims[0],))
        )
        if coords is None:
            # if dim is new then we need to assign it coordinates. If not
            # supplied, create them as list of integers ranging from 0 to
            # the number of input ntables
            coords = list(range(0, len(objs)))
        # if dim is new, we need to assign new coordinates which have either
        # been supplied by the caller or set to be a simple integer index
        new_dmap = new_dmap.assign_coords({dim: coords})

    new_dlist = []
    for dlist in dlists:
        new_dlist.extend(dlist)

    try:
        return NTable(new_dlist, new_dmap)
    except:
        new_dlist, new_dmap = handle_improper_broadcast(new_dlist, new_dmap)
        return NTable(new_dlist, new_dmap, engine=new_engine, ttype=new_ttype)


def default_refmap(*shape):
    dims = tuple(f"dim{i}" for i in range(len(shape)))
    coords = {dim: [] for dim in dims}
    for dim, size in zip(dims, shape):
        coords[dim] = [f"coord{i}" for i in range(size)]

    refarray = np.arange(np.prod(shape)).reshape(shape)
    refmap = xr.DataArray(refarray, coords, dims)
    return refmap


# class _Shell:
#     def __init__(self, obj):
#         self._obj = obj
#
#     def __TAPR_PEEL__(self):
#         return self._obj
#
#
# def shell(obj, layers=1):
#     return _Shell(obj)


def expand(iterable):
    result = []
    for item in iterable:
        try:
            result.append(expand(item))
        except TypeError:
            result.append(item)

    return result


def flatten(iterable):
    result = []
    for item in iterable:
        try:
            result.extend(flatten(item))
        except TypeError:
            result.append(item)

    return result


class _H5PYGroupDictifier(MutableMapping):
    def __init__(self, group=None):
        self._group = group

    def __getitem__(self, key):
        return self._group[key]

    def __setitem__(self, key, value):
        if isinstance(value, _H5PYGroupDictifier):
            if value._group is None:
                # if group is None, assume that we are meant to
                # just create a new child group with the name of
                # key.
                self._group.create_group(key)
            else:
                self._group[key] = value._group
        else:
            self._group[key] = value

    def __delitem__(self, key):
        raise NotImplementedError

    def __iter__(self):
        return iter(self._group)

    def __len__(self):
        return len(self._group)


def dictifyh5(group_or_file):
    if isinstance(group_or_file, h5py.File):
        return _H5PYGroupDictifier(group_or_file["/"])
    elif isinstance(group_or_file, h5py.Group):
        return _H5PYGroupDictifier(group_or_file)
    else:
        raise TypeError


# class Counter:
#     def __init__(self, start=0, stop=None, step=1):
#         self._start = start
#         self._stop = stop
#         self._step = step
#         self._current = start
#
#     def __call__(self, *args, **kwargs):
#         next_ = self._current + self._step
#         if self._stop is not None:
#             if next_ >= self._stop:
#                 raise ValueError("counter exceeded stop value")
#         self._current = next_
#         return self._current


class NTableStopIteration(Exception):
    pass



import unittest

import numpy as np

import xarray as xr

from tapr.main.conversion import ntable
from tapr.main.utils import (
    validate_engine,
    validate_ttype,
    validate_ntable_init,
    any_ntables,
    NULL,
    concatenate_ntables,
    default_refmap,
)
from tapr.main.engines import StandardEngine, ProcessEngine, ThreadEngine

from tests.testing_utils import assert_ntable_equivalent


class BadEngine:
    def __init__(self):
        pass

    def __call__(self, func, *args):
        return ["bad result"]


class TestValidationUtils(unittest.TestCase):
    def test_validate_engine_standard(self):
        standard_engine = StandardEngine()
        try:
            validate_engine(standard_engine)
        except ValueError as e:
            self.fail(f"validation failed due to exception:\n{e}")

    def test_validate_engine_thread(self):
        thread_engine = ThreadEngine(8)
        try:
            validate_engine(thread_engine)
        except ValueError as e:
            self.fail(f"validation failed due to exception:\n{e}")

    def test_validate_engine_process(self):
        process_engine = ProcessEngine(8)
        try:
            validate_engine(process_engine)
        except ValueError as e:
            self.fail(f"validation failed due to exception:\n{e}")

    def test_validate_engine_bad(self):
        bad_engine = BadEngine()
        try:
            validate_engine(bad_engine)
        except ValueError:
            pass
        except Exception:
            self.fail("unexpected exception was raised")
        else:
            self.fail("An exception was not raised")

    def test_validate_ttype_good(self):
        good_ttype = {np.ndarray, np.int32, int, float, object}
        try:
            validate_ttype(good_ttype)
        except ValueError as e:
            self.fail(f"validation failed due to exception:\n{e}")

    def test_validate_ttype_bad(self):
        bad_ttype = {np.ndarray, np.int64, int, float, object, "string"}
        try:
            validate_ttype(bad_ttype)
        except ValueError:
            pass
        except Exception:
            self.fail("unexpected exception was raised")
        else:
            self.fail("An exception was not raised")

    def test_validate_ntable_init_good(self):
        reflist = [1, 2, 3, 4, 5, 6]
        refmap = xr.DataArray(
            np.arange(6).reshape(2, 3),
            dims=("dim1", "dim2"),
            coords={"dim1": ["x1", "x2"], "dim2": ["y1", "y2", "y3"]},
        )
        engine = StandardEngine()
        ttype = {str, int, float}
        try:
            validate_ntable_init(reflist, refmap, engine, ttype)
        except Exception as e:
            self.fail(f"validation failed due to exception:\n{e}")

    def test_validate_ntable_init_bad(self):
        # refmap has invalid index for reflist
        reflist1 = [1, 2, 3, 4, 5]
        refmap1 = xr.DataArray(
            np.arange(6).reshape(2, 3),
            dims=("dim1", "dim2"),
            coords={"dim1": ["x1", "x2"], "dim2": ["y1", "y2", "y3"]},
        )
        engine = StandardEngine()
        ttype = {str, int, float}
        try:
            validate_ntable_init(reflist1, refmap1, engine, ttype)
        except ValueError:
            pass
        except Exception:
            self.fail("unexpected exception was raised")
        else:
            self.fail("An exception was not raised")

        # refmap has empty coordinates
        reflist2 = [1, 2, 3, 4, 5, 6]
        refmap2 = xr.DataArray(
            np.arange(6).reshape(2, 3),
        )
        engine = StandardEngine()
        ttype = {str, int, float}
        try:
            validate_ntable_init(reflist2, refmap2, engine, ttype)
        except ValueError:
            pass
        except Exception:
            self.fail("unexpected exception was raised")
        else:
            self.fail("An exception was not raised")

        # reflist is not a list
        reflist3 = (1, 2, 3, 4, 5, 6)
        refmap3 = xr.DataArray(
            np.arange(6).reshape(2, 3),
            dims=("dim1", "dim2"),
            coords={"dim1": ["x1", "x2"], "dim2": ["y1", "y2", "y3"]},
        )
        engine = StandardEngine()
        ttype = {str, int, float}
        try:
            validate_ntable_init(reflist3, refmap3, engine, ttype)
        except TypeError:
            pass
        except Exception:
            self.fail("unexpected exception was raised")
        else:
            self.fail("An exception was not raised")

        # refmap is not is not a DataArray
        reflist4 = [1, 2, 3, 4, 5, 6]
        refmap4 = np.arange(6).reshape(2, 3)
        engine = StandardEngine()
        ttype = {str, int, float}
        try:
            validate_ntable_init(reflist4, refmap4, engine, ttype)
        except TypeError:
            pass
        except Exception:
            self.fail("unexpected exception was raised")
        else:
            self.fail("An exception was not raised")

    def test_str_ntable_element(self):
        ntbl = ntable(
            {
                "x1": {
                    "y1": np.arange(6),
                    "y2": xr.DataArray(np.arange(6)),
                    "y3": 2,
                },
                "x2": {
                    "y1": "three_four_five_six_seven_eight_nine_ten",
                    "y2": [1, 2, 3],
                    "y3": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                },
            }
        )
        expected = """dim1                  y1                     y2                y3
dim0                                                             
x1    ndarray,(6,),int64  data array,(6,),int64                 2
x2         three...e_ten              [1, 2, 3]  [1, 2]...[9, 10]
Coordinates:
  * dim0     (dim0) <U2 'x1' 'x2'
  * dim1     (dim1) <U2 'y1' 'y2' 'y3'
Engine:
Standard (serial) Engine
Ttype:
DataArray|int|list|ndarray|str"""
        result = str(ntbl)
        self.assertEqual(result, expected)

    def test_str_ntable(self):
        dictionary = {}
        for i in range(150):
            dictionary[f"row{i}"] = {}
            for j in range(30):
                dictionary[f"row{i}"][f"col{j}"] = {}
                for k in range(15):
                    dictionary[f"row{i}"][f"col{j}"][f"page{k}"] = np.arange(
                        10
                    )
        ntbl = ntable(dictionary)
        # big boi
        result = str(ntbl)
        expected = """page0:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page1:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page2:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page3:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page4:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

. . .

###############################################################################

page10:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page11:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page12:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page13:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page14:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################


Coordinates:
  * dim0     (dim0) <U6 'row0' 'row1' 'row2' ... 'row147' 'row148' 'row149'
  * dim1     (dim1) <U5 'col0' 'col1' 'col2' 'col3' ... 'col27' 'col28' 'col29'
  * dim2     (dim2) <U6 'page0' 'page1' 'page2' ... 'page12' 'page13' 'page14'
Engine:
Standard (serial) Engine
Ttype:
ndarray"""
        self.assertEqual(result, expected)

        # 0 dim ntbl
        result = str(ntbl.struct[0, 0, 0])
        expected = """ndarray,(10,),int64
Coordinates:
    dim0     <U6 'row0'
    dim1     <U5 'col0'
    dim2     <U6 'page0'
Engine:
Standard (serial) Engine
Ttype:
ndarray"""
        self.assertEqual(result, expected)

        # smol boi
        result = str(ntbl.dim2.page0.struct[0:5, 0:3])
        expected = """dim1                 col0                 col1                 col2
dim0                                                               
row0  ndarray,(10,),int64  ndarray,(10,),int64  ndarray,(10,),int64
row1  ndarray,(10,),int64  ndarray,(10,),int64  ndarray,(10,),int64
row2  ndarray,(10,),int64  ndarray,(10,),int64  ndarray,(10,),int64
row3  ndarray,(10,),int64  ndarray,(10,),int64  ndarray,(10,),int64
row4  ndarray,(10,),int64  ndarray,(10,),int64  ndarray,(10,),int64
Coordinates:
  * dim0     (dim0) <U6 'row0' 'row1' 'row2' 'row3' 'row4'
  * dim1     (dim1) <U5 'col0' 'col1' 'col2'
    dim2     <U6 'page0'
Engine:
Standard (serial) Engine
Ttype:
ndarray"""

        self.assertEqual(result, expected)

        # biggish boi
        result = str(ntbl.struct[:, :, 0:5])
        # print(result)
        expected = """page0:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page1:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page2:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page3:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################

page4:

dim1             col0           col1           col2           col3  ...          col26          col27          col28          col29
dim0                                                                ...                                                            
row0    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row1    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row2    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row3    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row4    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row5    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row6    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row7    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row8    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row9    ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row140  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row141  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row142  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row143  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row144  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row145  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row146  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row147  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row148  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64
row149  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64  ...  ndarr...int64  ndarr...int64  ndarr...int64  ndarr...int64

[20 rows x 30 columns]

###############################################################################


Coordinates:
  * dim0     (dim0) <U6 'row0' 'row1' 'row2' ... 'row147' 'row148' 'row149'
  * dim1     (dim1) <U5 'col0' 'col1' 'col2' 'col3' ... 'col27' 'col28' 'col29'
  * dim2     (dim2) <U6 'page0' 'page1' 'page2' 'page3' 'page4'
Engine:
Standard (serial) Engine
Ttype:
ndarray"""

        self.assertEqual(result, expected)

    def test_any_ntables(self):
        ntbl = ntable(
            {
                "x1": {"y1": 0, "y2": "1", "y3": 2},
                "x2": {"y1": "three", "y2": 4, "y3": "5"},
            }
        )
        self.assertTrue(any_ntables((0, 1, 2, ntbl, 0, ntbl)))

    def test_null(self):
        null = NULL()
        self.assertEqual("NULL", str(null))
        self.assertEqual("NULL", repr(null))
        self.assertEqual("NULL", null.__ntable_element__str__())
        self.assertTrue(isinstance(null[0], NULL))
        self.assertTrue(isinstance(null(), NULL))
        self.assertTrue(isinstance(null + 10, NULL))

    def test_concat_ntables(self):
        dictionary = {
            "x1": {"y1": 0, "y2": "1", "y3": 2},
            "x2": {"y1": "three", "y2": 4, "y3": "5"},
        }
        ntbl = ntable(dictionary)
        new_ntbl = concatenate_ntables((ntbl, ntbl), dim="dim2")
        expected_ntbl = ntable(
            {0: dictionary, 1: dictionary}, dims=("dim2", "dim0", "dim1")
        ).struct.transpose("dim0", "dim1", "dim2")

        assert_ntable_equivalent(new_ntbl, expected_ntbl)

        dictionary1 = {
            "x1": {"y1": 0, "y2": "1", "y3": NULL()},
            "x2": {"y1": "three", "y2": 4, "y3": NULL()},
        }

        new_ntbl1 = concatenate_ntables((ntbl, ntbl.struct[:,0:2]), dim="dim2")

        expected_ntbl1 = ntable(
            {0: dictionary, 1: dictionary1}, dims=("dim2", "dim0", "dim1")
        ).struct.transpose("dim0", "dim1", "dim2")

        assert_ntable_equivalent(new_ntbl1, expected_ntbl1)

    def test_default_refmap(self):
        refmap = default_refmap(3,2)
        expected = xr.DataArray(np.arange(6).reshape((3,2)), coords={"dim0":["coord0","coord1","coord2"], "dim1":["coord0","coord1"]})
        self.assertTrue(refmap.equals(expected))


if __name__ == "__main__":
    unittest.main()



import unittest


from tapr.main.conversion import ntable
from tapr.main.engines import StandardEngine, ThreadEngine, ProcessEngine


def func(a, b):
    return a + b


class TestStandardEngine(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )
        self._ntbl_b = ntable(
            {
                "row1": {"col1": "a", "col2": "b"},
                "row2": {"col1": "c", "col2": "d"},
            }
        )

    def test_call(self):
        engine = StandardEngine()
        result = engine.__tapr_engine__map__(func, self._ntbl_a.reflist, self._ntbl_b.reflist)
        self.assertListEqual(
            result,
            [
                item1 + item2
                for item1, item2 in zip(
                    self._ntbl_a.reflist, self._ntbl_b.reflist
                )
            ],
        )


class TestThreadEngine(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )
        self._ntbl_b = ntable(
            {
                "row1": {"col1": "a", "col2": "b"},
                "row2": {"col1": "c", "col2": "d"},
            }
        )

    def test_call(self):
        engine = ThreadEngine(threads=8)
        result = engine.__tapr_engine__map__(func, self._ntbl_a.reflist, self._ntbl_b.reflist)
        self.assertListEqual(
            result,
            [
                item1 + item2
                for item1, item2 in zip(
                    self._ntbl_a.reflist, self._ntbl_b.reflist
                )
            ],
        )


class TestProcessEngine(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )
        self._ntbl_b = ntable(
            {
                "row1": {"col1": "a", "col2": "b"},
                "row2": {"col1": "c", "col2": "d"},
            }
        )

    def test_call(self):
        engine = ProcessEngine(processes=8)
        result = engine.__tapr_engine__map__(func, self._ntbl_a.reflist, self._ntbl_b.reflist)
        self.assertListEqual(
            result,
            [
                item1 + item2
                for item1, item2 in zip(
                    self._ntbl_a.reflist, self._ntbl_b.reflist
                )
            ],
        )


if __name__ == "__main__":
    unittest.main()



from .conversion import ntable, tabulate
from .engines import StandardEngine, ProcessEngine, ThreadEngine
from .ntable import NTable
from .tabularization import tabularize
from .handling import handled_by, FunctionError
from .filtering import contains, matches
from .utils import NULL



from collections import namedtuple
import unittest

from dataclasses import dataclass

import numpy as np
import pandas as pd
import xarray as xr
import operator as op

from tapr.main.ntable import NTable
from tapr.main.conversion import ntable
from tapr.main.handling import FunctionError
from tapr.main.alchemy import NTableMapAlchemy, NTableAlchemy
from tests.testing_utils import assert_ntable_equivalent


def aaa():
    return "aaa"


@dataclass
class Test:
    a: int


class TestNTableCreation(unittest.TestCase):
    def setUp(self):
        self._reflist_a = [0, "1", 2, "three", 4, "5"]
        self._refmap_a = xr.DataArray(
            np.arange(6).reshape((2, 3)),
            coords={"dim_0": ["x1", "x2"], "dim_1": ["y1", "y2", "y3"]},
            dims=["dim_0", "dim_1"],
        )
        self._reflist_b = [
            0,
            "1",
            2,
            "three",
            4,
            "5",
            6,
            "sev",
            8.0,
            9,
            10,
            11,
        ]
        self._refmap_b = xr.DataArray(
            np.arange(12).reshape((3, 4)),
            coords={
                "dim_0": ["x1", "x2", "x3"],
                "dim_1": ["y1", "y2", "y3", "y4"],
            },
            dims=["dim_0", "dim_1"],
        )

    def test_init(self):
        # test good init
        ntbl = NTable(self._reflist_a, self._refmap_a)
        # test mismatch reflist/refmap
        self.assertRaises(ValueError, NTable, self._reflist_a, self._refmap_b)


class TestNTableCoversion(unittest.TestCase):
    def setUp(self):
        self._reflist_a = [0, "1", 2, "three", 4, "5"]
        self._refmap_a = xr.DataArray(
            np.arange(6).reshape((2, 3)),
            coords={"dim_0": ["x1", "x2"], "dim_1": ["y1", "y2", "y3"]},
            dims=["dim_0", "dim_1"],
        )
        self._ntable_a = NTable(self._reflist_a, self._refmap_a)

    def test_to_dictionary(self):
        dictionary = self._ntable_a.to_dictionary()
        self.assertDictEqual(
            dictionary,
            {
                "x1": {"y1": 0, "y2": "1", "y3": 2},
                "x2": {"y1": "three", "y2": 4, "y3": "5"},
            },
        )

    def test_to_pandas(self):
        dframe = self._ntable_a.to_pandas()
        test_dframe = pd.DataFrame(
            {
                "x1": {"y1": 0, "y2": "1", "y3": 2},
                "x2": {"y1": "three", "y2": 4, "y3": "5"},
            }
        ).T
        test_dframe.index.rename("dim_0", inplace=True)
        test_dframe.columns.rename("dim_1", inplace=True)
        pd.util.testing.assert_frame_equal(dframe, test_dframe)

    def test_to_data_array(self):
        darray = self._ntable_a.to_data_array()
        test_darray = xr.DataArray(
            np.array([[0, "1", 2], ["three", 4, "5"]], dtype="object"),
            coords={"dim_0": ["x1", "x2"], "dim_1": ["y1", "y2", "y3"]},
            dims=["dim_0", "dim_1"],
        )
        xr.testing.assert_equal(darray, test_darray)


class TestNTableCore(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )

        self._ntbl_b = ntable(
            {
                "row1": {"col1": list("r1c1"), "col2": list("r1c2")},
                "row2": {"col1": list("r2c1"), "col2": list("r2c2")},
            }
        )

        self._ntbl_c = ntable(
            {
                "row1": {"col1": aaa, "col2": aaa},
                "row2": {"col1": aaa, "col2": aaa},
            }
        )

        self._ntbl_d = ntable(
            {
                "row1": {"col1": np.arange(3), "col2": np.arange(3)},
                "row2": {"col1": np.arange(3), "col2": np.arange(3)},
            }
        )

    def test_iter(self):
        expected0 = ntable(
            {
                "row1": {"col1": "r", "col2": "r"},
                "row2": {"col1": "r", "col2": "r"},
            }
        )
        expected1 = ntable(
            {
                "row1": {"col1": "1", "col2": "1"},
                "row2": {"col1": "2", "col2": "2"},
            }
        )
        expected2 = ntable(
            {
                "row1": {"col1": "c", "col2": "c"},
                "row2": {"col1": "c", "col2": "c"},
            }
        )
        expected3 = ntable(
            {
                "row1": {"col1": "1", "col2": "2"},
                "row2": {"col1": "1", "col2": "2"},
            }
        )
        expecteds = [expected0, expected1, expected2, expected3]
        for result, expected in zip(self._ntbl_a, expecteds):
            assert_ntable_equivalent(result, expected)

    def test_getitem(self):
        # in bounds
        expected = ntable(
            {
                "row1": {"col1": "r", "col2": "r"},
                "row2": {"col1": "r", "col2": "r"},
            }
        )
        result = self._ntbl_a[0]
        assert_ntable_equivalent(result, expected)

        # out of bounds
        expected = ntable(
            {
                "row1": {
                    "col1": FunctionError(
                        IndexError("string index out of range"),
                        op.getitem,
                        "r1c1",
                        4,
                    ),
                    "col2": FunctionError(
                        IndexError("string index out of range"),
                        op.getitem,
                        "r1c2",
                        4,
                    ),
                },
                "row2": {
                    "col1": FunctionError(
                        IndexError("string index out of range"),
                        op.getitem,
                        "r2c1",
                        4,
                    ),
                    "col2": FunctionError(
                        IndexError("string index out of range"),
                        op.getitem,
                        "r2c2",
                        4,
                    ),
                },
            }
        )
        result = self._ntbl_a[4]
        assert_ntable_equivalent(result, expected)

    def test_setitem(self):
        self._ntbl_b[1] = "lol"
        expected = ntable(
            {
                "row1": {
                    "col1": ["r", "lol", "c", "1"],
                    "col2": ["r", "lol", "c", "2"],
                },
                "row2": {
                    "col1": ["r", "lol", "c", "1"],
                    "col2": ["r", "lol", "c", "2"],
                },
            }
        )
        assert_ntable_equivalent(self._ntbl_b, expected)

    def test_call(self):
        expected = ntable(
            {
                "row1": {"col1": "aaa", "col2": "aaa"},
                "row2": {"col1": "aaa", "col2": "aaa"},
            }
        )
        result = self._ntbl_c()
        assert_ntable_equivalent(result, expected)

    def test_array_ufunc(self):
        expected = ntable(
            {
                "row1": {
                    "col1": np.sin(np.arange(3)),
                    "col2": np.sin(np.arange(3)),
                },
                "row2": {
                    "col1": np.sin(np.arange(3)),
                    "col2": np.sin(np.arange(3)),
                },
            }
        )
        result = np.sin((self._ntbl_d))
        self.assertTrue(all(np.array_equal(result, expected).struct.flat))

    def test_array_function(self):
        expected = ntable(
            {
                "row1": {
                    "col1": np.vstack((np.arange(3), np.arange(3))),
                    "col2": np.vstack((np.arange(3), np.arange(3))),
                },
                "row2": {
                    "col1": np.vstack((np.arange(3), np.arange(3))),
                    "col2": np.vstack((np.arange(3), np.arange(3))),
                },
            }
        )
        result = np.vstack((self._ntbl_d, self._ntbl_d))
        self.assertTrue(all(np.array_equal(result, expected).struct.flat))

    def test_getattr(self):
        result = self._ntbl_d.shape
        expected = ntable(
            {
                "row1": {"col1": (3,), "col2": (3,)},
                "row2": {"col1": (3,), "col2": (3,)},
            }
        )
        assert_ntable_equivalent(result, expected)

    def test_str(self):
        result = str(self._ntbl_a)
        expected = """dim1    col1    col2
dim0                
row1  "r1c1"  "r1c2"
row2  "r2c1"  "r2c2"
Coordinates:
  * dim0     (dim0) <U4 'row1' 'row2'
  * dim1     (dim1) <U4 'col1' 'col2'
Engine:
Standard (serial) Engine
Ttype:
str"""
        self.assertEqual(result, expected)


class TestNTableMap(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )

    def test_setitem_old_key(self):
        dim0_map = self._ntbl_a.ntable_map("dim0")
        dim0_map["row1"] = dim0_map["row2"]
        expected = ntable(
            {
                "row1": {"col1": "r2c1", "col2": "r2c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )
        assert_ntable_equivalent(self._ntbl_a, expected)

    def test_setitem_new_key(self):
        dim0_map = self._ntbl_a.ntable_map("dim0")
        dim0_map["row3"] = dim0_map["row2"]
        expected = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
                "row3": {"col1": "r2c1", "col2": "r2c2"},
            }
        )
        assert_ntable_equivalent(self._ntbl_a, expected)

    def test_setitem_list_key(self):
        insert = ntable(
            {
                "row1": {"col1": "a", "col2": "b"},
                "row2": {"col1": "c", "col2": "d"},
            }
        )
        dim0_map = self._ntbl_a.ntable_map("dim0")
        dim0_map[["row1", "row2"]] = insert
        expected = ntable(
            {
                "row1": {"col1": "a", "col2": "b"},
                "row2": {"col1": "c", "col2": "d"},
            }
        )
        assert_ntable_equivalent(self._ntbl_a, expected)

    def test_setitem_list_key_invalid_value_length(self):
        insert = ntable(
            {
                "row1": {"col1": "a", "col2": "b"},
                "row2": {"col1": "c", "col2": "d"},
            }
        )
        dim0_map = self._ntbl_a.ntable_map("dim0")
        try:
            dim0_map[["row1"]] = insert
        except ValueError:
            self.assertTrue(True)

    def test_setitem_list_key_scalar_value(self):
        dim0_map = self._ntbl_a.ntable_map("dim0")
        dim0_map[["row3", "row4"]] = 3
        expected = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
                "row3": {"col1": 3, "col2": 3},
                "row4": {"col1": 3, "col2": 3},
            }
        )
        assert_ntable_equivalent(self._ntbl_a, expected)


class TestTabularizedAttributes(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": Test(1), "col2": Test(2)},
                "row2": {"col1": Test(3), "col2": Test(4)},
            }
        )

    def test_getattr(self):
        result = self._ntbl_a.tattr.a
        expected = ntable(
            {
                "row1": {"col1": 1, "col2": 2},
                "row2": {"col1": 3, "col2": 4},
            }
        )
        assert_ntable_equivalent(result, expected)

    def test_setattr(self):
        self._ntbl_a.tattr.a = 10
        expected = ntable(
            {
                "row1": {"col1": Test(10), "col2": Test(10)},
                "row2": {"col1": Test(10), "col2": Test(10)},
            }
        )
        assert_ntable_equivalent(self._ntbl_a, expected)


class TestMisc(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": Test(1), "col2": Test(2)},
                "row2": {"col1": Test(3), "col2": Test(4)},
            }
        )

    def test_ntable_map_ntable(self):
        result = self._ntbl_a.dim0.ntable
        self.assertTrue(result is self._ntbl_a)

    def test_ntable_map_alchemy(self):
        result = self._ntbl_a.dim0.alchemy
        self.assertTrue(isinstance(result, NTableMapAlchemy))

    def test_ntable_map_str(self):
        result = str(self._ntbl_a.dim0)
        expected = """{'row1': dim1
col1    Test
col2    Test
dtype: object
Coordinates:
    dim0     <U4 'row1'
  * dim1     (dim1) <U4 'col1' 'col2'
Engine:
Standard (serial) Engine
Ttype:
Test, 'row2': dim1
col1    Test
col2    Test
dtype: object
Coordinates:
    dim0     <U4 'row2'
  * dim1     (dim1) <U4 'col1' 'col2'
Engine:
Standard (serial) Engine
Ttype:
Test}"""
        self.assertEqual(result, expected)

    def test_ntable_map_repr(self):
        result = repr(self._ntbl_a.dim0)
        expected = """{'row1': dim1
col1    Test
col2    Test
dtype: object
Coordinates:
    dim0     <U4 'row1'
  * dim1     (dim1) <U4 'col1' 'col2'
Engine:
Standard (serial) Engine
Ttype:
Test, 'row2': dim1
col1    Test
col2    Test
dtype: object
Coordinates:
    dim0     <U4 'row2'
  * dim1     (dim1) <U4 'col1' 'col2'
Engine:
Standard (serial) Engine
Ttype:
Test}"""
        self.assertEqual(result, expected)

    def test_ntable_map_delitem(self):
        try:
            del self._ntbl_a.dim0["row1"]
        except NotImplementedError:
            self.assertTrue(True)

    def test_ntable_map_contains(self):
        result = self._ntbl_a.dim0.contains("1")
        expected = ntable(
            {
                "row1": {"col1": Test(1), "col2": Test(2)},
            }
        )
        assert_ntable_equivalent(result, expected)

    def test_ntable_map_matches(self):
        result = self._ntbl_a.dim0.matches(r"r.*2")
        expected = ntable(
            {
                "row2": {"col1": Test(3), "col2": Test(4)},
            }
        )
        assert_ntable_equivalent(result, expected)

    def test_ntable_map_relabel(self):
        result = self._ntbl_a.dim0.relabel(row1="row3", row2="row4")
        expected = ntable(
            {
                "row3": {"col1": Test(1), "col2": Test(2)},
                "row4": {"col1": Test(3), "col2": Test(4)},
            }
        )
        assert_ntable_equivalent(result, expected)

    def test_ntable_alchemy(self):
        result = self._ntbl_a.alchemy
        self.assertTrue(isinstance(result, NTableAlchemy))

    def test_ntable_getattr_err(self):
        result = False
        try:
            self._ntbl_a.spacedoge
        except AttributeError:
            result = True

        self.assertTrue(result)

    def test_ntable_item(self):
        result = self._ntbl_a.struct.loc["row1", "col1"].item()
        expected = Test(1)
        self.assertEqual(result, expected)

    def test_ntable_to_pandas_error(self):
        ntbl = ntable(
            {
                "page1": {
                    "row3": {"col1": Test(1), "col2": Test(2)},
                    "row4": {"col1": Test(3), "col2": Test(4)},
                }
            }
        )
        result = False
        try:
            blah = ntbl.to_pandas()
        except ValueError:
            result = True

        self.assertTrue(result)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from tapr.main.conversion import ntable
from tapr.main.engines import ProcessEngine, ThreadEngine
from tapr.io.ntableio import save_ntable, load_ntable
from tests.testing_utils import assert_ntable_equivalent


class Test(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": 3.0, "col2": 3, "col3": (3,), "col4":set([3,]), "col5": complex(3,3)},
                "row2": {"col1": np.array(3), "col2": "3", "col3":[3,], "col4":b"3", "col5":bytearray(b"3")},
            }
        )
        self._ntbl_b = ntable(
            {
                "row1": {"col1": 3.0, "col2": 3, "col3": (3,), "col4":set([3,]), "col5": complex(3,3)},
                "row2": {"col1": np.array(3), "col2": "3", "col3":[3,], "col4":b"3", "col5":bytearray(b"3")},
            }, engine=ThreadEngine(8)
        )
        self._ntbl_c = ntable(
            {
                "row1": {"col1": 3.0, "col2": 3, "col3": (3,), "col4":set([3,]), "col5": complex(3,3)},
                "row2": {"col1": np.array(3), "col2": "3", "col3":[3,], "col4":b"3", "col5":bytearray(b"3")},
            }, engine=ProcessEngine(8)
        )

    def test_save_and_load(self):
        save_ntable(self._ntbl_a, "/tmp/test_a.ntbl")
        result = load_ntable("/tmp/test_a.ntbl")
        assert_ntable_equivalent(result, self._ntbl_a)

        # thread engine
        save_ntable(self._ntbl_b, "/tmp/test_b.ntbl")
        result = load_ntable("/tmp/test_b.ntbl")
        assert_ntable_equivalent(result, self._ntbl_b)

        # process engine
        save_ntable(self._ntbl_c, "/tmp/test_c.ntbl")
        result = load_ntable("/tmp/test_c.ntbl")
        assert_ntable_equivalent(result, self._ntbl_c)







if __name__ == "__main__":
    unittest.main()


import unittest

import numpy as np
import pandas as pd
import xarray as xr


from tapr.main.processing import broadcast_tables, tabular_map
from tapr.main.conversion import ntable
from tapr.main.ntable import NTable
from tapr.main.utils import NULL
from tapr.main.engines import StandardEngine
from tests.testing_utils import assert_ntable_equivalent


class TestBroadcastTables(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )
        self._ntbl_b = ntable({"col1": "c1", "col2": "c2"}, dims=("dim1",))
        self._ntbl_c = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
                "row3": {"col1": "r3c1", "col2": "r3c2"},
            }
        )

    def test_broadcast_tables_good(self):
        ntbl1, ntbl2 = broadcast_tables(self._ntbl_a, self._ntbl_b)
        assert_ntable_equivalent(ntbl1, self._ntbl_a)
        reflist = ["c1", "c2", "c1", "c2"]
        test_ntbl2 = NTable(reflist, self._ntbl_a.refmap)
        assert_ntable_equivalent(ntbl2, test_ntbl2)

    def test_broadcast_tables_bad(self):
        ntbl1, ntbl2 = broadcast_tables(self._ntbl_a, self._ntbl_c)
        test_ntbl1 = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
                "row3": {"col1": NULL(), "col2": NULL()},
            }
        )
        assert_ntable_equivalent(ntbl1, test_ntbl1)
        assert_ntable_equivalent(ntbl2, self._ntbl_c)


class TestTabularMap(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )
        self._ntbl_b = ntable({"col1": "c1", "col2": "c2"}, dims=("dim1",))
        # self._ntbl_c = ntable(
        #     {
        #         "row1": {"col1": "r1c1", "col2": "r1c2"},
        #         "row2": {"col1": "r2c1", "col2": "r2c2"},
        #         "row3": {"col1": "r3c1", "col2": "r3c2"},
        #     }
        # )

    def test_tabular_map_good(self):
        def func(a, b):
            return a + b

        result = tabular_map(
            (func, StandardEngine()), self._ntbl_a, self._ntbl_a
        )

        expected_ntbl = ntable(
            {
                "row1": {"col1": "r1c1r1c1", "col2": "r1c2r1c2"},
                "row2": {"col1": "r2c1r2c1", "col2": "r2c2r2c2"},
            }
        )
        assert_ntable_equivalent(result, expected_ntbl)

    def test_tabular_map_bad(self):
        def func(a, b):
            return a + b

        try:
            result = tabular_map(
                (func, StandardEngine()), self._ntbl_a, self._ntbl_b
            )
        except ValueError:
            pass
        except Exception as e:
            self.fail(f"Unexpected exception {e} was raised")
        else:
            self.fail("ExpectedException not raised")


if __name__ == "__main__":
    unittest.main()



import re







def isin(collection):
    def function(value):
        return value in collection

    return function


def contains(string):
    def function(value):
        return string in value

    return function


def matches(pattern):
    def function(value):
        return bool(re.match(pattern, value))

    return function


class NTableFilter:
    def __init__(self, ntbl):
        self._ntbl = ntbl

    def __getitem__(self, index):
        ntbl = self._ntbl
        for k, v in index.items():
            # v should be a function that returns True or False based on some
            # condition on the items of the iterable (in this case the keys of
            # the dimension mapping)
            findex = list(filter(v, ntbl.ntable_map(k)))
            ntbl = ntbl.ntable_map(k)[findex]

        return ntbl

    def __call__(self, **kwargs):
        return self[kwargs]



from tapr.main.utils import NULL

def assert_ntable_equivalent(ntbl1, ntbl2):
    reflist1 = ntbl1.reflist
    refmap1 = ntbl1.refmap
    reflist2 = ntbl2.reflist
    refmap2 = ntbl2.refmap
    for dim1, dim2 in zip(refmap1.dims, refmap2.dims):
        if (refmap1.coords[dim1] != refmap2.coords[dim2]).any():
            raise ValueError("NTable objects are not equivalent")
    for index1, index2 in zip(refmap1.values.flat, refmap2.values.flat):
        value1 = reflist1[index1]
        value2 = reflist2[index2]
        if value1 != value2:
            # check to see if values are NULL, which are falsey
            if isinstance(value1, NULL) and isinstance(value2, NULL):
                continue
            raise ValueError("NTable objects are not equivalent")


import itertools as it

import numpy as np
import xarray as xr

from .utils import basic_refmap, full, full_lite, handle_improper_broadcast
from .engines import StandardEngine


def broadcast_tables(*args, lite=False):
    from .ntable import NTable

    ntbls = [arg for arg in args if isinstance(arg, NTable)]
    nons = [arg for arg in args if not isinstance(arg, NTable)]
    # ensure there is at least one ntable object
    if len(ntbls) == 0:
        raise ValueError("args must contain at least one NTable object")

    dmaps = [ntbl.refmap for ntbl in ntbls]
    dlists = [ntbl.reflist for ntbl in ntbls]
    engines = [ntbl.engine for ntbl in ntbls]
    ttypes = [ntbl.ttype for ntbl in ntbls]
    new_dmaps = xr.broadcast(*dmaps)
    new_ntbls = []
    for dlist, dmap, engine, ttype in zip(dlists, new_dmaps, engines, ttypes):
        try:
            new_ntbls.append(NTable(dlist, dmap, engine, ttype))
        except:
            # if the above fails, perhaps it was because an improper broadcast
            # took place. Try handling it.
            dlist, dmap = handle_improper_broadcast(dlist, dmap)
            new_ntbls.append(NTable(dlist, dmap, engine, ttype))

    if lite:
        non_ntbls = [
            full_lite(non, new_dmaps[0].coords, new_dmaps[0].dims)
            for non in nons
        ]
    else:
        non_ntbls = [
            full(non, new_dmaps[0].coords, new_dmaps[0].dims) for non in nons
        ]

    # return results in the right order
    result = []
    for item in args:
        if isinstance(item, NTable):
            result.append(new_ntbls.pop(0))
        else:
            result.append(non_ntbls.pop(0))

    return result


def tabular_map(func_engine, *ntable_args):
    from .ntable import NTable

    if isinstance(func_engine, tuple):
        func = func_engine[0]
        engine = func_engine[1]
    else:
        func = func_engine
        engine = StandardEngine()
    new_reflist = list(
        engine.__tapr_engine__map__(func, *(ntbl.struct.flat for ntbl in ntable_args))
    )
    new_refmap = basic_refmap(
        ntable_args[0].refmap.coords, ntable_args[0].refmap.dims
    )

    # TODO: Change validate to False once new_reflist and new_refmap
    # consistency is proven to be guaranteed
    result_engine = ntable_args[0].engine
    return NTable(new_reflist, new_refmap, result_engine, validate=True)



import operator as op
import mmap

import numpy as np
import xarray as xr
import h5py

from ..main.utils import xarray_coords_to_dict


def save_ntable(ntbl, fname, allow_pickle=False):
    """
    Save NTable object to a file.

    Parameters
    ----------
    ntbl : NTable
        The NTable to save.
    fname : str
        The name of the file to save it as.

    Returns
    -------
    None.

    """
    from .serialization import serialize
    from ..main.tabularization import tabularize

    bytes_ntable, type_id_ntable = tabularize()(serialize)(ntbl, allow_pickle=allow_pickle)
    len_ntable = tabularize()(len)(bytes_ntable)

    engine_bytes, engine_type_id = serialize(ntbl.engine)

    total_bytes = bytearray()
    for bytes_ in bytes_ntable.struct.flat:
        total_bytes += bytes_

    total_bytes = total_bytes + engine_bytes

    len_array = len_ntable.to_data_array(dtype=int)

    stop_array = xr.DataArray(
        np.cumsum(len_array.values).reshape(len_array.shape),
        coords=len_array.coords,
        dims=len_array.dims,
    )
    start_array = stop_array - len_array
    type_id_array = type_id_ntable.to_data_array()

    ublock_size = 1 << (len(total_bytes) - 1).bit_length()
    if ublock_size < 512:
        ublock_size = 512
    coords_dict = xarray_coords_to_dict(ntbl.struct.coords)
    with h5py.File(fname, "w", userblock_size=ublock_size) as fo:
        fo["/start"] = start_array.values
        fo["/stop"] = stop_array.values
        fo["/type"] = type_id_array.values
        fo.attrs["engine_length"] = len(engine_bytes)
        fo.attrs["engine_type_id"] = engine_type_id
        for i, dim in enumerate(ntbl.struct.dims):
            labels = coords_dict[dim]
            fo[f"/coords/{dim}"] = np.string_(labels)
            fo[f"/coords/{dim}"].attrs["axis"] = i

    with open(fname, "r+b") as fo:
        fo.write(total_bytes)


def load_ntable(fname, filter={}, allow_pickle=False):
    """

    Parameters
    ----------
    fname : str
        The name file to load.

    Returns
    -------
    loaded_ntable : NTable
        The loaded NTable object.

    """
    from .serialization import deserialize
    from ..main.conversion import ntable
    from ..main.tabularization import tabularize

    with h5py.File(fname, "r") as fo:
        userblock_size = fo.userblock_size
        start_ndarray = fo["/start"][...]
        stop_ndarray = fo["/stop"][...]
        type_id_ndarray = fo["/type"].asstr()[...]
        data_stop_loc = stop_ndarray.max()
        engine_length = fo.attrs["engine_length"]
        engine_type_id = fo.attrs["engine_type_id"]
        dims = []
        axes = []
        coords_dict = {}
        for dim, coord_dset in fo["/coords/"].items():
            dims.append(dim)
            axes.append(coord_dset.attrs["axis"])
            coords_dict[dim] = [
                item.decode() for item in list(coord_dset[...])
            ]

        dims = [dim for _, dim in sorted(zip(axes, dims))]

    with open(fname, "rb") as fo:
        total_bytes = mmap.mmap(fo.fileno(),userblock_size, access=mmap.ACCESS_READ)
        engine_bytes = total_bytes[
            data_stop_loc : data_stop_loc + engine_length
        ]

    start_ntable = ntable(xr.DataArray(start_ndarray, coords_dict, dims)).filter[filter]
    stop_ntable = ntable(xr.DataArray(stop_ndarray, coords_dict, dims)).filter[filter]
    type_id_ntable = ntable(xr.DataArray(type_id_ndarray, coords_dict, dims)).filter[filter]

    bytes_ntable = tabularize()(op.getitem)(
        total_bytes, slice(start_ntable, stop_ntable)
    )

    loaded_ntable = tabularize()(deserialize)(bytes_ntable, type_id_ntable, allow_pickle=allow_pickle)
    loaded_ntable.engine = deserialize(engine_bytes, engine_type_id)
    return loaded_ntable



import itertools as it

import numpy as np

from .ntable import NTable
from .tabularization import tabularize
from .utils import full, NULL, default_refmap, xarray_coords_to_dict, basic_refmap


def blank(coords, dims, engine=None, ttype=None):
    """
    Creates an NTable with given coordinates and dimensions
    whose elements are all NULL()

    Parameters
    ----------
    coords:
        The coordinates for the desired blank NTable
    dims:
        The dimensions for the desired blank NTable
    engine:
        The engine for the resulting NTable to use. Default is None.
    ttype:
        The ttype for the resulting NTable to use. Default is None.

    Returns
    -------
    ntable (NTable): The desired NTable

    """
    return full(NULL(), coords=coords, dims=dims, engine=engine, ttype=ttype)


def sblank(*shape, engine=None, ttype=None):
    """
    Similar to blank, but based on a given shape instead of
    coordinates and dimensions.

    Parameters
    ----------
    shape : tuple
        The shape for the desired blank NTable
    engine:
        The engine for the resulting NTable to use. Default is None.
    ttype:
        The ttype for the resulting NTable to use. Default is None.

    Returns
    -------
    ntable (NTable): The desired NTable

    """
    total_size = 1
    for size in shape:
        total_size *= size

    reflist = [NULL()] * total_size
    refmap = default_refmap(*shape)
    return NTable(reflist, refmap, engine=engine, ttype=ttype)


def cartograph(ntbl, engine=None, ttype=None):
    """
    Creates an NTable whose elements are the intersection (tuple) of the
    corresponding coordinates of the input NTable

    Parameters
    ----------
    ntbl: NTable
        The NTable to base the resulting NTable on
    engine:
        The engine for the resulting NTable to use. Default is None.
    ttype:
        The ttype for the resulting NTable to use. Default is None.

    Returns
    -------
    ntable (NTable): The desired NTable

    """
    coords = ntbl.struct.coords
    dims = ntbl.struct.dims
    coords_dict = xarray_coords_to_dict(coords, dims)
    reflist = list(it.product(*coords_dict.values()))
    refmap = basic_refmap(coords, dims)
    return NTable(reflist, refmap, engine=engine, ttype=ttype)

def count(ntbl, engine=None, ttype=None):
    coords = ntbl.struct.coords
    dims = ntbl.struct.dims
    coords_dict = xarray_coords_to_dict(coords, dims)
    reflist = list(i for i,_ in enumerate(it.product(*coords_dict.values())))
    refmap = ntbl.refmap.copy()
    return NTable(reflist, refmap, engine=engine, ttype=ttype)


# def itertable(iterable, engine=None, ttype=None):
#     """
#     Converts potentially nested iterables into a representative
#     NTable.
#
#     Parameters
#     ----------
#     iterable: iterable
#         the iterable to generate the NTable from
#     engine:
#         The engine for the resulting NTable to use. Default is None.
#     ttype:
#         The ttype for the resulting NTable to use. Default is None.
#
#     Returns
#     -------
#     ntable: NTable
#         The desired NTable
#
#     """
#     expanded = expand(iterable)
#     refarray = np.array(expanded, dtype="object")
#     reflist = []
#     for item in refarray.flat:
#         try:
#             # see if the item was wrapped in a shell
#             reflist.append(item.__TAPR_PEEL__())
#         except AttributeError:
#             reflist.append(item)
#
#     refmap = default_refmap(*refarray.shape)
#     return NTable(reflist, refmap, engine=engine, ttype=ttype)



import unittest

from tapr.main.conversion import ntable
from tapr.main.alchemy import NTableMapAlchemy, NTableAlchemy
from tests.testing_utils import assert_ntable_equivalent


class TestNTableAlchemy(unittest.TestCase):
    def setUp(self):
        self._ntbl_a = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
            }
        )

    def test_getitem(self):
        result = self._ntbl_a.dim0.alchemy[r"r(.*)1"]
        expected = ntable(
            {
                "r__ALCHEMY__1": {"col1": "r1c1", "col2": "r1c2"},
            }
        )
        assert_ntable_equivalent(result, expected)

    def test_setitem(self):
        self._ntbl_a.dim0.alchemy["3"] = self._ntbl_a.dim0.alchemy[r"r.*(1)"]
        expected = ntable(
            {
                "row1": {"col1": "r1c1", "col2": "r1c2"},
                "row2": {"col1": "r2c1", "col2": "r2c2"},
                "row3": {"col1": "r1c1", "col2": "r1c2"},
            }
        )
        assert_ntable_equivalent(self._ntbl_a, expected)


if __name__ == "__main__":
    unittest.main()



import functools as ft

from .processing import tabular_map, broadcast_tables
from .utils import any_ntables, call_args_kwargs
from .conversion import tabulate
from .engines import Engine, StandardEngine


class _Tabularized:
    def __init__(self, func, engine):
        self._func = func
        self._engine = engine

    def __call__(self, *args, **kwargs):
        from .ntable import NTable

        try:
            targs = tabulate(args)
        except ValueError:
            targs = args
        try:
            tkwargs = tabulate(kwargs)
        except ValueError:
            tkwargs = kwargs
        if not isinstance(targs, NTable) and not isinstance(tkwargs, NTable):
            # if there are no NTable objects in args or kwargs, just call the
            # function normally on the inputs
            return self._func(*args, **kwargs)

        bfunc, bargs, bkwargs = broadcast_tables(
            self._func, targs, tkwargs, lite=True
        )
        # bfunc engine will always be just a StandardEngine since it is
        # generated via a (scalar, NTable, NTable or scalar) broadcast. We want
        # the resulting engine to be that of bargs. This can be assured by
        # setting the engine of the first argument to tabular_map (bfunc) to
        # be that of bargs. Same for ttype being STANDARD_TTYPE
        bfunc.engine = bargs.engine
        bfunc.ttype = bargs.ttype
        return tabular_map(
            (call_args_kwargs, self._engine), bfunc, bargs, bkwargs
        )

    def __str__(self):
        return f"Tabularized:\nfunc: {self._func.__name__}\nengine: {self._engine}"

    def __repr__(self):
        return str(self)

def tabularize(engine=None):
    if engine is None:
        engine = StandardEngine()

    if not isinstance(engine, Engine):
        raise TypeError(f"engine must be of type Engine")

    def tabulizer(func):
        if callable(func):
            return _Tabularized(func, engine)
        else:
            raise TypeError(f"func must be callable, which {func} is not")

    return tabulizer



import pathlib as pl
from setuptools import setup, find_packages

HERE = pl.Path(__file__).parent

README_PATH = HERE / "README.md"

with open(README_PATH, "r") as rm:
    README_TEXT = rm.read()

setup(
    name="tapr",
    version="0.2.0",
    description="A library to facilitate tabular programming.",
    long_description=README_TEXT,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/elspacedoge/tapr",
    author="elspacedoge",
    author_email="elspacedoge@gmail.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent"
    ],
    package_dir={"": "./"},
    packages=find_packages(where="./"),
    include_package_data=True,
    install_requires=["numpy", "pandas", "xarray", "h5py", "matplotlib", "plotly"]
)


import warnings as wn
import itertools as it

import numpy as np
import xarray as xr

from .utils import concatenate_ntables, xarray_coords_to_dict, basic_refmap


class _LocIndexer:
    """
    Index by labels rather than position.
    """

    def __init__(self, struct):
        self._struct = struct

    def __getitem__(self, index):
        from .ntable import NTable

        new_refmap = self._struct.ntable.refmap.loc[index]
        return NTable(self._struct.ntable.reflist, new_refmap, validate=False)

    def __setitem__(self, index, value):
        from .ntable import NTable

        index_map = self._struct.ntable.refmap.loc[index]

        if len(np.unique(index_map)) < index_map.values.size:
            wn.warn(
                """Warning: The index corresponds to a part of the
                    structure that is multi-referential. This means that there 
                    are multiple items in the refmap that point to the same 
                    item in the reflist. As such, assignment operations could result in some
                    unexpected behavior."""
            )

        def _setreflist(i, v):
            self._struct.ntable.ttype.add(type(v))
            self._struct.ntable.reflist[i] = v

        if not isinstance(value, NTable):
            list(
                map(
                    _setreflist,
                    index_map.values.flat,
                    (value for item in index_map.values.flat),
                )
            )
        else:
            # should numpy or xarray broadcasting be used here?
            bi, bv = xr.broadcast(index_map, value.refmap)
            if np.isnan(bi.values).any() or np.isnan(bv.values).any():
                raise ValueError("Unable to assign input value")
            list(
                map(
                    _setreflist,
                    bi.values.flat,
                    (value.reflist[vi] for vi in bv.values.flat),
                )
            )


class NTableStructure:
    """
    NTable structure object. Used for structure-oriented operations such as
    indexing, assigning, transposing, iterating over elements, etc.

    Parameters
    ----------
    ntbl : NTable
        The ntable object whose structure is to be worked with.

    """

    def __init__(self, ntbl):
        self._ntbl = ntbl

    @property
    def ntable(self):
        """The corresponding NTable object."""
        return self._ntbl

    @property
    def flat(self):
        """An iterable that flattens the structure of the NTable object."""
        return (self._ntbl.reflist[i] for i in self._ntbl.refmap.values.flat)

    @property
    def dims(self):
        """The dimensions of the NTable object"""
        return self._ntbl.refmap.dims

    @property
    def coords(self):
        """The coords of the NTable object"""
        return self._ntbl.refmap.coords

    @property
    def name(self):
        """The name of the NTable object"""
        return self._ntbl.refmap.name

    @property
    def ndim(self):
        """The number of dimensions in the NTable object"""
        return self._ntbl.refmap.ndim

    @property
    def shape(self):
        """The shape of the NTable object"""
        return self._ntbl.refmap.shape

    @property
    def size(self):
        """The size of the NTable object"""
        return self._ntbl.refmap.size

    @property
    def loc(self):
        """Coordinate name based indexer"""
        return _LocIndexer(self)

    @property
    def T(self):
        """The transposed NTable object."""
        from .ntable import NTable

        return NTable(
            self._ntbl.reflist,
            self._ntbl.refmap.T,
            self._ntbl.engine,
            self._ntbl.ttype,
        )

    def __getitem__(self, index):
        from .ntable import NTable

        new_refmap = self._ntbl.refmap[index]
        return NTable(self._ntbl.reflist, new_refmap)

    def __setitem__(self, index, value):
        from .ntable import NTable

        index_map = self._ntbl.refmap[index]

        if len(np.unique(index_map)) < index_map.values.size:
            wn.warn(
                """Warning: The index corresponds to a part of the
                    structure that is multi-referential. This means that there 
                    are multiple items in the refmap that point to the same 
                    item in the reflist. As such, assignment operations could result in some
                    unexpected behavior."""
            )

        def _setreflist(i, v):
            self._ntbl.ttype.add(type(v))
            self._ntbl.reflist[i] = v

        if not isinstance(value, NTable):
            list(
                map(
                    _setreflist,
                    index_map.values.flat,
                    (value for item in index_map.values.flat),
                )
            )
        else:
            # should numpy or xarray broadcasting be used here?
            bi, bv = xr.broadcast(index_map, value.refmap)
            if np.isnan(bi.values).any() or np.isnan(bv.values).any():
                raise ValueError("Unable to assign input value")
            list(
                map(
                    _setreflist,
                    bi.values.flat,
                    (value.reflist[vi] for vi in bv.values.flat),
                )
            )

    def __add__(self, right):
        return concatenate_ntables(
            (self.ntable, right.ntable), dim=self.dims[0]
        )

    def transpose(self, *refmap_args, **refmap_kwargs):
        from .ntable import NTable

        return NTable(
            self._ntbl.reflist,
            self._ntbl.refmap.transpose(*refmap_args, **refmap_kwargs),
            self._ntbl.engine,
            self._ntbl.ttype,
        )

    def item(self):
        """
        If the NTable object has just a single element (regardless of its
        dimensionality), returns that item.
        """
        if self.size > 1:
            raise ValueError(
                f"Cannot return a single item for NTable Structure of size greater than 1 ({self.size})"
            )
        return next(self.flat)

    def relabel(self, **kwargs):
        """
        Relabel coordinates.

        Parameters
        ----------
        **kwargs : dictionary-like
            Keys specify which dimension to apply the relabel to, values
            should be mappings themselves that map old labels along the
            corresponding dimension to new labels.

        Returns
        -------
        NTable
            The relabled NTable.

        """
        from .ntable import NTable

        refmap = self._ntbl.refmap
        for k, v in kwargs.items():
            refmap = refmap.to_dataset(k).rename(**v).to_array(k)

        return NTable(self._ntbl.reflist, refmap)

    def flatter(self, dims=None):
        if dims is None:
            dims = tuple()
        for dim in dims:
            if dim not in self.dims:
                raise ValueError(f"dim {dim} not found")
        dims = tuple(dim for dim in self.dims if dim not in dims)
        coord_dict = xarray_coords_to_dict(self.coords)
        indexes = it.product(*(coord_dict[dim] for dim in dims))
        for index in indexes:
            yield self.loc[{dim: coord for dim, coord in zip(dims, index)}]

    def compress(self, dims=None):
        from .ntable import NTable

        new_reflist = [ntbl for ntbl in self.flatter(dims)]
        new_dims = tuple(dim for dim in self.dims if dim not in dims)
        coords_dict = xarray_coords_to_dict(self.coords)
        new_coords = {dim: coords_dict[dim] for dim in new_dims}

        new_refmap = basic_refmap(new_coords, new_dims)

        return NTable(new_reflist, new_refmap, engine=self.ntable.engine)



import re

from .filtering import matches

ALCHEMY_STRING = "__ALCHEMY__"


class NTableAlchemy:
    def __init__(self, ntbl):
        self._ntbl = ntbl

    def __getitem__(self, index):
        ntbl = self._ntbl
        for k, v in index.items():
            # k should be a dimension name, v should be a regex
            findex = list(filter(matches(v), ntbl.ntable_map(k)))
            ntbl = ntbl.ntable_map(k)[findex]
            label_map = {}
            for label in ntbl.ntable_map(k):
                match = re.match(v, label)
                if len(match.groups()) == 0:
                    label_map[label] = label.replace(
                        match.string, ALCHEMY_STRING
                    )
                else:
                    new_label = label
                    for string in match.groups():
                        new_label = new_label.replace(string, ALCHEMY_STRING)
                    label_map[label] = new_label
            ntbl = ntbl.ntable_map(k).relabel(**label_map)

        return ntbl

    def __setitem__(self, index, value):
        from .ntable import NTable

        if not isinstance(value, NTable):
            raise ValueError(
                f"value must be of type NTable, not {type(value)}"
            )

        for k, v in index.items():
            my_ntable_map = self._ntbl.ntable_map(k)
            ur_ntable_map = value.ntable_map(k)
            alchemy_labels = list(
                ur_ntable_map.contains(ALCHEMY_STRING).ntable_map(k)
            )
            label_map = {
                label: label.replace(ALCHEMY_STRING, v)
                for label in alchemy_labels
            }

            for old_label, new_label in label_map.items():
                my_ntable_map[new_label] = ur_ntable_map[old_label]


class NTableMapAlchemy:
    def __init__(self, ntable_map):
        self._ntable_map = ntable_map

    def __getitem__(self, index):
        return NTableAlchemy(self._ntable_map.ntable)[
            {self._ntable_map.dim: index}
        ]

    def __setitem__(self, index, value):
        NTableAlchemy(self._ntable_map.ntable)[
            {self._ntable_map.dim: index}
        ] = value



import importlib as il
alchemy = il.import_module(".main.alchemy", package=__name__)
conversion = il.import_module(".main.conversion", package=__name__)
defs = il.import_module(".main.defs", package=__name__)
engines = il.import_module(".main.engines", package=__name__)
filtering = il.import_module(".main.filtering", package=__name__)
handling = il.import_module(".main.handling", package=__name__)
ntable = il.import_module(".main.ntable", package=__name__)
processing = il.import_module(".main.processing", package=__name__)
qol = il.import_module(".main.qol", package=__name__)
structure = il.import_module(".main.structure", package=__name__)
tabularization = il.import_module(".main.tabularization", package=__name__)
ttypes = il.import_module(".main.ttypes", package=__name__)
utils = il.import_module(".main.utils", package=__name__)

io = il.import_module(".io", package=__name__)

from .main.conversion import ntable
from .main.qol import blank, sblank, cartograph, count
from .main.tabularization import tabularize
from .main.utils import full, full_lite, full_like, concatenate_ntables as concatenate

from .io.ntableio import save_ntable, load_ntable

_px = il.import_module(".visualization.plotly.express", package=__name__)
# look for any tabularized plotly express functions and expose them here
for name in dir(_px):
    value = getattr(_px,name)
    if isinstance(value, tabularization._Tabularized):
        globals()[name] = value